﻿/***************************************************************************

Copyright (c) Microsoft Corporation. All rights reserved.
This code is licensed under the Visual Studio SDK license terms.
THIS CODE IS PROVIDED *AS IS* WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING ANY
IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR
PURPOSE, MERCHANTABILITY, OR NON-INFRINGEMENT.

***************************************************************************/

using System;
using System.Diagnostics;
using System.Linq;
using System.Xml;
using System.Xml.Linq;
using Microsoft.AnalysisServices;

namespace Microsoft.DaxEditor
{
    public class DaxCalcProperty
    {
        public enum FormatType
        {
            General,
            NumberDecimal,
            NumberWhole,
            Percentage,
            Scientific,
            Currency,
            DateTimeCustom,
        }

        public FormatType Format { get; set; }
        public int? Accuracy { get; set; }
        public bool? ThousandSeparator { get; set; }
        public string FormatString { get; set; }
        public string CalculationType { get; set; }
        public string CustomFormat { get; set; }

        private DaxCalcProperty()
        {
        }

        public static DaxCalcProperty CreateDefaultCalculationProperty()
        {
            var rv = new DaxCalcProperty();
            rv.FormatString = "''";
            rv.CalculationType = "Member";
            return rv;
        }

        public static DaxCalcProperty CreateFromXElement(XElement cp)
        {
            var rv = new DaxCalcProperty();

            if (cp != null)
            {
                Debug.Assert(cp.Name.LocalName == "CalculationProperty");
                var annotations = cp.Element(MeasuresContainer.NS + "Annotations");
                rv.FormatString = cp.Element(MeasuresContainer.NS + "FormatString").Value;
                rv.CalculationType = cp.Element(MeasuresContainer.NS + "CalculationType").Value;

                var formatAnnotationElement = annotations.Elements(MeasuresContainer.NS + "Annotation").First(i => string.Equals(i.Element(MeasuresContainer.NS + "Name").Value, "Format"));
                if (formatAnnotationElement != null)
                {
                    var formatValueElement = formatAnnotationElement.Element(MeasuresContainer.NS + "Value").Element("Format");
                    Debug.Assert(formatValueElement != null);
                    Debug.Assert(formatValueElement.Name == "Format");
                    var formatString = formatValueElement.Attribute("Format").Value;
                    rv.Format = (FormatType)Enum.Parse(typeof(FormatType), formatString);
                    switch (rv.Format)
                    {
                        case FormatType.General:
                        case FormatType.NumberWhole:
                        case FormatType.NumberDecimal:
                        case FormatType.Percentage:
                        case FormatType.Scientific:
                        case FormatType.Currency:
                        case FormatType.DateTimeCustom:
                            var accuracyAttribute = formatValueElement.Attribute("Accuracy");
                            if (accuracyAttribute != null)
                                rv.Accuracy = int.Parse(accuracyAttribute.Value);

                            var thousandSeparatorAttribute = formatValueElement.Attribute("ThousandSeparator");
                            if (thousandSeparatorAttribute != null)
                                rv.ThousandSeparator = bool.Parse(thousandSeparatorAttribute.Value);

                            if (formatValueElement.FirstNode != null)
                                rv.CustomFormat = InitCustomFormatString((XElement)formatValueElement.FirstNode);

                            break;

                        default:
                            Debug.Assert(false, "Not reachable");
                            break;
                    }
                }
            }
            else
            {
                rv.FormatString = "''";
                rv.CalculationType = "Member";
            }

            return rv;
        }

        private static string InitCustomFormatString(XElement customFormatElement)
        {
            Debug.Assert(customFormatElement != null);
            Debug.Assert(customFormatElement.Name.LocalName == "Currency" || customFormatElement.Name.LocalName == "DateTimes");
            if (customFormatElement.Name.LocalName == "DateTimes")
                customFormatElement = customFormatElement.Element("DateTime");

            string customFormat = "'";
            customFormat += string.Join(" ", customFormatElement.Attributes().Select(i => string.Format(@"{0}=""{1}""", i.Name, i.Value)));
            customFormat += "'";
            return customFormat;
        }

        public string ToDax()
        {
            string result;

            switch (Format)
            {
                case FormatType.General:
                    result = string.Empty;
                    break;

                case FormatType.NumberDecimal:
                case FormatType.NumberWhole:
                case FormatType.Percentage:
                case FormatType.Scientific:
                case FormatType.Currency:
                case FormatType.DateTimeCustom:
                    result = "CALCULATION PROPERTY " + Format.ToString();
                    if (Accuracy.HasValue)
                        result += " Accuracy=" + Accuracy.Value;

                    if (ThousandSeparator.HasValue && ThousandSeparator.Value == true)
                        result += " ThousandSeparator=True";

                    if (!string.IsNullOrEmpty(FormatString) && !string.Equals("''", FormatString))
                        result += " Format=" + FormatString;

                    if (!string.IsNullOrEmpty(CustomFormat))
                        result += " AdditionalInfo=" + CustomFormat;
                    break;
                default:
                    Debug.Assert(false, "Not reachable");
                    result = string.Empty;
                    break;
            }

            return result;
        }

        public void ToXml(XmlWriter writer, string measureName)
        {
            writer.WriteStartElement("CalculationProperty");
            writer.WriteStartElement("Annotations");

            writer.WriteStartElement("Annotation");
            writer.WriteElementString("Name", "Type");
            writer.WriteElementString("Value", "User");
            writer.WriteEndElement(); // Annotation

            writer.WriteStartElement("Annotation");
            writer.WriteElementString("Name", "IsPrivate");
            writer.WriteElementString("Value", "False");
            writer.WriteEndElement(); // Annotation

            writer.WriteStartElement("Annotation");
            writer.WriteElementString("Name", "Format");

            writer.WriteStartElement("Value");

            writer.WriteStartElement("Format", "");
            writer.WriteAttributeString("Format", Format.ToString());
            if (Accuracy.HasValue)
                writer.WriteAttributeString("Accuracy", Accuracy.Value.ToString());
            if (ThousandSeparator.HasValue && ThousandSeparator.Value == true)
                writer.WriteAttributeString("ThousandSeparator", "True");

            if(!string.IsNullOrEmpty(CustomFormat))
            {
                if (Format == FormatType.Currency)
                {
                    writer.WriteRaw("<Currency ");
                    writer.WriteRaw(CustomFormat.Trim('\''));
                    writer.WriteRaw("/>");
                }
                else if (Format == FormatType.DateTimeCustom)
                {
                    writer.WriteStartElement("DateTimes");
                    writer.WriteRaw("<DateTime ");
                    writer.WriteRaw(CustomFormat.Trim('\''));
                    writer.WriteRaw("/>");
                    writer.WriteEndElement();
                }
                else
                {
                    Debug.Fail("Not reachable code");
                }
            }

            writer.WriteEndElement(); // Format
            
            writer.WriteEndElement(); // Value
            
            writer.WriteEndElement(); // Annotation

            writer.WriteEndElement(); // Annotations
            writer.WriteElementString("CalculationReference", measureName);
            writer.WriteElementString("CalculationType", "Member");
            writer.WriteElementString("FormatString", FormatString);

            writer.WriteEndElement(); // CalculationProperty
        }
    }
}
