﻿/***************************************************************************

Copyright (c) Microsoft Corporation. All rights reserved.

THIS CODE IS PROVIDED *AS IS* WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING ANY
IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR
PURPOSE, MERCHANTABILITY, OR NON-INFRINGEMENT.

***************************************************************************/

using System;
using System.Diagnostics;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using Microsoft.VisualStudio.Text.Editor;
using System.Windows;
using System.Windows.Documents;
using System.Text;
using System.Windows.Threading;
using System.Data;
using System.Windows.Forms.Integration;

namespace Microsoft.DaxEditor
{
    class DaxEditorMargin : StackPanel, IWpfTextViewMargin, IUpdateEditorMargin
    {
        public const string MarginName = "DaxEditorMargin";
        private IWpfTextView _textView;
        private bool _isDisposed = false;
        private Border _splitter;
        private TabControl _tabControl;
        private TabItem _resultTab;
        private TabItem _schemaTab;
        private WebBrowser _schemaWebBrowser;
        private System.Windows.Forms.DataGridView _dataGridView;
        private WindowsFormsHost _host;

        /// <summary>
        /// Creates a <see cref="DaxEditorMargin"/> for a given <see cref="IWpfTextView"/>.
        /// </summary>
        /// <param name="textView">The <see cref="IWpfTextView"/> to attach the margin to.</param>
        public DaxEditorMargin(IWpfTextView textView)
        {
            _textView = textView;
            ClipToBounds = true;

            _splitter = new Border()
            {
                Height = 5,
                Background = new SolidColorBrush(Colors.Goldenrod),
                VerticalAlignment = System.Windows.VerticalAlignment.Top,
                Cursor = Cursors.SizeNS,
                ForceCursor = true,
            };

            _splitter.MouseMove += OnSplitterMouseMove;
            _splitter.MouseDown += OnSplitterMouseDown;
            _splitter.MouseUp   += OnSplitterMouseUp;
            _splitter.MouseLeave += OnSplitterMouseLeave;

            this.Children.Add(_splitter);

            _tabControl = new TabControl() { Height = 300 };
            _resultTab = new TabItem() { Header = "Result" };
            _schemaTab = new TabItem() { Header = "Schema" };

            _tabControl.Items.Add(_resultTab);
            _tabControl.Items.Add(_schemaTab);

            _dataGridView = new System.Windows.Forms.DataGridView()
            {
                ReadOnly = true,
                AllowUserToAddRows = false,
                AllowUserToDeleteRows = false,
                AllowUserToOrderColumns = true,
                AllowUserToResizeRows = true,
                AutoSizeColumnsMode = System.Windows.Forms.DataGridViewAutoSizeColumnsMode.DisplayedCells,
            };

            _host = new WindowsFormsHost();
            _host.Child = _dataGridView;

            _resultTab.Content = _host;

            _schemaWebBrowser = new WebBrowser();
            _schemaWebBrowser.Height = _host.Height;
            _schemaTab.Content = _schemaWebBrowser;

            this.Children.Add(_tabControl);
        }

        private void ThrowIfDisposed()
        {
            if (_isDisposed)
                throw new ObjectDisposedException(MarginName);
        }

        #region Splitter move support
        void OnSplitterMouseMove(object sender, MouseEventArgs e)
        {
            if (this._splitter.IsMouseCaptured && e.LeftButton == MouseButtonState.Pressed)
            {
                _tabControl.Height = Math.Max(ActualHeight - e.GetPosition(this).Y, 1);
            }
        }

        void OnSplitterMouseDown(object sender, MouseButtonEventArgs e)
        {
            if (sender == this._splitter)
            {
                Mouse.Capture(this._splitter);
            }
        }

        void OnSplitterMouseUp(object sender, MouseButtonEventArgs e)
        {
            if (this._splitter.IsMouseCaptured)
            {
                Mouse.Capture(null);
            }
        }

        void OnSplitterMouseLeave(object sender, MouseEventArgs e)
        {
            Mouse.OverrideCursor = null;
        }
        #endregion

        #region IWpfTextViewMargin Members

        /// <summary>
        /// The <see cref="Sytem.Windows.FrameworkElement"/> that implements the visual representation
        /// of the margin.
        /// </summary>
        public System.Windows.FrameworkElement VisualElement
        {
            // Since this margin implements Canvas, this is the object which renders
            // the margin.
            get
            {
                ThrowIfDisposed();
                return this;
            }
        }

        #endregion

        #region ITextViewMargin Members

        public double MarginSize
        {
            // Since this is a horizontal margin, its width will be bound to the width of the text view.
            // Therefore, its size is its height.
            get
            {
                ThrowIfDisposed();
                return this.ActualHeight;
            }
        }

        public bool Enabled
        {
            // The margin should always be enabled
            get
            {
                ThrowIfDisposed();
                return true;
            }
        }

        /// <summary>
        /// Returns an instance of the margin if this is the margin that has been requested.
        /// </summary>
        /// <param name="marginName">The name of the margin requested</param>
        /// <returns>An instance of EditorMarginVstsTest or null</returns>
        public ITextViewMargin GetTextViewMargin(string marginName)
        {
            return (marginName == DaxEditorMargin.MarginName) ? (IWpfTextViewMargin)this : null;
        }

        public void Dispose()
        {
            if (!_isDisposed)
            {
                GC.SuppressFinalize(this);
                _isDisposed = true;
            }
        }
        #endregion

        public void UpdateResult(DataTable result)
        {
            if (_dataGridView.InvokeRequired)
            {
                _dataGridView.Invoke(
                    new Action(
                        () =>
                        {
                            _dataGridView.DataSource = result;
                        }
                    )
                );
            }
            else
            {
                _dataGridView.DataSource = result;
            }
        }

        public void UpdateSchema(string schemaHtml)
        {
            if (!_schemaWebBrowser.Dispatcher.CheckAccess())
            {
                _schemaWebBrowser.Dispatcher.Invoke(
                    new Action(
                        () =>
                        {
                            _schemaWebBrowser.NavigateToString(schemaHtml);
                        }
                    )
                );
            }
            else
            {
                _schemaWebBrowser.NavigateToString(schemaHtml);
            }
        }
    }
}
