﻿/***************************************************************************

Copyright (c) Microsoft Corporation. All rights reserved.
This code is licensed under the Visual Studio SDK license terms.
THIS CODE IS PROVIDED *AS IS* WITHOUT WARRANTY OF
ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING ANY
IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR
PURPOSE, MERCHANTABILITY, OR NON-INFRINGEMENT.

***************************************************************************/

using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using Microsoft.VisualStudio.Package;
using System;

namespace Microsoft.DaxEditor
{
    public class MeasuresContainer
    {
        public const string NS = "{http://schemas.microsoft.com/analysisservices/2003/engine}";
        public const string NS200 = "{http://schemas.microsoft.com/analysisservices/2010/engine/200}";

        private MeasuresContainer(IList<DaxMeasure> measures)
        {
            Measures = measures;
        }

        private IList<DaxMeasure> _measures;

        public static MeasuresContainer ParseXmlaAsText(string xmlaText)
        {
            var alterScriptDoc = XDocument.Parse(xmlaText);
            var mdxScript = alterScriptDoc.Descendants(NS + "MdxScript");
            Debug.Assert(1 == mdxScript.Count());
            return ParseXmlaAsXElement(mdxScript.First());
        }

        public static MeasuresContainer ParseXmlaAsXElement(XElement mdxScript)
        {
            var allMeasures = new List<DaxMeasure>();
            foreach (var mdxCommand in mdxScript.Element(NS + "Commands").Elements(NS + "Command"))
            {
                var commandText = mdxCommand.Element(NS + "Text").Value;
                var measures = ParseDaxScript(commandText);
                foreach (var measure in measures.Measures)
                {
                    var calcProperty = mdxScript.Element(NS + "CalculationProperties").Elements().FirstOrDefault(i => i.Element(NS + "CalculationReference").Value == measure.NameInBrackets);
                    measure.CalcProperty = DaxCalcProperty.CreateFromXElement(calcProperty);
                    allMeasures.Add(measure);
                }
            }

            return new MeasuresContainer(allMeasures);
        }

        /// <summary>
        /// Parses measures from the mdx script text.  If there is an error during parsing an exception is thrown.
        /// </summary>
        /// <param name="daxScript">mdx script text</param>
        /// <returns>Enumerable dax measures</returns>
        public static MeasuresContainer ParseDaxScript(string daxScript)
        {
            var handler = new Babel.Parser.ErrorHandler();
            var scanner = new Babel.Lexer.Scanner();
            var parser = new Babel.Parser.Parser();  // use noarg constructor
            parser.scanner = scanner;
            scanner.Handler = handler;
            parser.SetHandler(handler);
            scanner.SetSourceText(daxScript);

            var request = new ParseRequest(true);
            request.Sink = new AuthoringSink(ParseReason.None, 0, 0, 100);
            parser.MBWInit(request);
            var parseResult = parser.Parse();
            if (handler.Errors)
                throw DaxParsingException.FromHandler(handler);

            return new MeasuresContainer(parser.Measures);
        }

        public IList<DaxMeasure> Measures
        {
            get { return _measures; }
            private set { _measures = value; }
        }

        public string GetDaxText()
        {
            var sb = new StringBuilder();
            foreach (var measure in Measures)
            {
                var measureText = measure.FullText;
                sb.Append(measure.FullText);
                var calcPropertyDax = measure.CalcProperty.ToDax();
                if (!string.IsNullOrEmpty(calcPropertyDax))
                {
                    sb.AppendLine();
                    sb.Append(calcPropertyDax);
                }
                sb.Append(';');
                sb.AppendLine();
                sb.AppendLine();
            }
            sb.AppendLine();

            return sb.ToString();
        }

        /// <summary>
        /// Update input XMLA with the measures of this object.
        /// </summary>
        /// <param name="inputXmla">input XMLA</param>
        /// <returns>updated XMLA with measures from this object</returns>
        public string UpdateMeasuresInXmla(string inputXmla)
        {
            if (string.IsNullOrEmpty(inputXmla))
                throw new ArgumentException("input");

            var serverCommandProducer = new ServerCommandProducer(inputXmla);
            var newMdxScript = serverCommandProducer.ProduceAlterScriptElement(Measures);
            var inputDocument = XDocument.Parse(inputXmla);

            var actualMdxScriptDocument = XDocument.Parse(newMdxScript);
            inputDocument.Descendants(MeasuresContainer.NS + "MdxScript").First().ReplaceWith(actualMdxScriptDocument.Root);

            return inputDocument.ToString(SaveOptions.OmitDuplicateNamespaces);
        }
    }
}
